@@ -1,8 +1,11 @@
 # -*- coding: utf-8 -*-
 
+from thread import start_new_thread, get_ident
 import threading
 import unittest
+import time
 import sys
+import gc
 
 import lupa
 
@@ -14,6 +17,10 @@ except NameError:
     def _next(o):
         return o.next()
 
+def _wait():
+    # A crude wait/yield function not relying on synchronization primitives.
+    time.sleep(0.01)
+
 class SetupLuaRuntimeMixin(object):
 
     def setUp(self):
@@ -21,7 +28,7 @@ class SetupLuaRuntimeMixin(object):
 
     def tearDown(self):
         self.lua = None
-
+        gc.collect()
 
 class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_eval(self):
@@ -476,6 +483,39 @@ class TestLuaCoroutines(SetupLuaRuntimeMixin, unittest.TestCase):
         gen = co(pyfunc, 5)
         self.assertEqual([0,1,0,1,0,1], list(gen))
 
+    def test_coroutine_send(self):
+        lua_code = '''\
+            function()
+                local i = 0
+                while coroutine.yield(i) do i = i+1 end
+                return i   -- not i+1 !
+            end
+        '''
+        count = self.lua.eval(lua_code).coroutine()
+        result = [ count.send(value) for value in ([None]+[True]*9+[False]) ]
+        self.assertEqual(list(range(10)) + [9], result)
+        self.assertRaises(StopIteration, count.send, True)
+
+    def test_coroutine_send_with_arguments(self):
+        lua_code = '''\
+            function(N)
+                local i = 0
+                while coroutine.yield(i) < N do i = i+1 end
+                return i   -- not i+1 !
+            end
+        '''
+        count = self.lua.eval(lua_code).coroutine(5)
+        result = []
+        try:
+            for value in ([None]+list(range(10))):
+                result.append( count.send(value) )
+        except StopIteration:
+            pass
+        else:
+            self.assertTrue(False)
+        self.assertEqual(list(range(6)) + [5], result)
+        self.assertRaises(StopIteration, count.send, True)
+
     def test_coroutine_status(self):
         lua_code = '''\
         coroutine.create(
@@ -548,6 +588,9 @@ class TestLuaCoroutines(SetupLuaRuntimeMixin, unittest.TestCase):
 
 
 class TestLuaApplications(unittest.TestCase):
+    def tearDown(self):
+        gc.collect()
+
     def test_mandelbrot(self):
         # copied from Computer Language Benchmarks Game
         code = '''\
@@ -596,6 +639,9 @@ end
 
 
 class TestLuaRuntimeEncoding(unittest.TestCase):
+    def tearDown(self):
+        gc.collect()
+
     unicode_type = type(IS_PYTHON3 and 'abc' or 'abc'.decode('ASCII'))
 
     test_string = '"abcüöä"'
@@ -632,6 +678,8 @@ class TestLuaRuntimeEncoding(unittest.TestCase):
 
 
 class TestMultipleLuaRuntimes(unittest.TestCase):
+    def tearDown(self):
+        gc.collect()
 
     def test_multiple_runtimes(self):
         lua1 = lupa.LuaRuntime()
@@ -663,10 +711,15 @@ class TestMultipleLuaRuntimes(unittest.TestCase):
 
 
 class TestThreading(unittest.TestCase):
+    def tearDown(self):
+        gc.collect()
 
-    def _run_threads(self, threads):
+    def _run_threads(self, threads, starter=None):
         for thread in threads:
             thread.start()
+        if starter is not None:
+            time.sleep(0.1) # give some time to start up
+            starter.set()
         for thread in threads:
             thread.join()
 
@@ -684,16 +737,18 @@ class TestThreading(unittest.TestCase):
         functions = [ lua.execute(func_code) for _ in range(10) ]
         results = [None] * len(functions)
 
+        starter = threading.Event()
         def test(i, func, *args):
+            starter.wait()
             results[i] = func(*args)
 
-        threads = [ threading.Thread(target=test, args=(i, func, 20))
+        threads = [ threading.Thread(target=test, args=(i, func, 25))
                     for i, func in enumerate(functions) ]
 
-        self._run_threads(threads)
+        self._run_threads(threads, starter)
 
         self.assertEqual(1, len(set(results)))
-        self.assertEqual(13529, results[0])
+        self.assertEqual(150049, results[0])
 
     def test_threading(self):
         func_code = '''\
@@ -864,6 +919,230 @@ class TestThreading(unittest.TestCase):
         ##     image = Image.fromstring('1', (image_size, image_size), result_bytes)
         ##     image.show()
 
+
+
+################################################################################
+# tests for the FastRLock implementation
+
+class TestFastRLock(unittest.TestCase):
+    """Copied from CPython's test.lock_tests module
+    """
+    def setUp(self):
+        from lupa._lupa import FastRLock
+        self.locktype = FastRLock
+
+    def tearDown(self):
+        gc.collect()
+
+    class Bunch(object):
+        """
+        A bunch of threads.
+        """
+        def __init__(self, f, n, wait_before_exit=False):
+            """
+            Construct a bunch of `n` threads running the same function `f`.
+            If `wait_before_exit` is True, the threads won't terminate until
+            do_finish() is called.
+            """
+            self.f = f
+            self.n = n
+            self.started = []
+            self.finished = []
+            self._can_exit = not wait_before_exit
+            def task():
+                tid = get_ident()
+                self.started.append(tid)
+                try:
+                    f()
+                finally:
+                    self.finished.append(tid)
+                    while not self._can_exit:
+                        _wait()
+            for i in range(n):
+                start_new_thread(task, ())
+
+        def wait_for_started(self):
+            while len(self.started) < self.n:
+                _wait()
+
+        def wait_for_finished(self):
+            while len(self.finished) < self.n:
+                _wait()
+
+        def do_finish(self):
+            self._can_exit = True
+
+
+    # the locking tests
+
+    """
+    Tests for both recursive and non-recursive locks.
+    """
+
+    def test_constructor(self):
+        lock = self.locktype()
+        del lock
+
+    def test_acquire_destroy(self):
+        lock = self.locktype()
+        lock.acquire()
+        del lock
+
+    def test_acquire_release(self):
+        lock = self.locktype()
+        lock.acquire()
+        lock.release()
+        del lock
+
+    def test_try_acquire(self):
+        lock = self.locktype()
+        self.assertTrue(lock.acquire(False))
+        lock.release()
+
+    def test_try_acquire_contended(self):
+        lock = self.locktype()
+        lock.acquire()
+        result = []
+        def f():
+            result.append(lock.acquire(False))
+        self.Bunch(f, 1).wait_for_finished()
+        self.assertFalse(result[0])
+        lock.release()
+
+    def test_acquire_contended(self):
+        lock = self.locktype()
+        lock.acquire()
+        N = 5
+        def f():
+            lock.acquire()
+            lock.release()
+
+        b = self.Bunch(f, N)
+        b.wait_for_started()
+        _wait()
+        self.assertEqual(len(b.finished), 0)
+        lock.release()
+        b.wait_for_finished()
+        self.assertEqual(len(b.finished), N)
+
+    def test_with(self):
+        lock = self.locktype()
+        def f():
+            lock.acquire()
+            lock.release()
+        def _with(err=None):
+            with lock:
+                if err is not None:
+                    raise err
+        _with()
+        # Check the lock is unacquired
+        self.Bunch(f, 1).wait_for_finished()
+        self.assertRaises(TypeError, _with, TypeError)
+        # Check the lock is unacquired
+        self.Bunch(f, 1).wait_for_finished()
+
+    def test_thread_leak(self):
+        # The lock shouldn't leak a Thread instance when used from a foreign
+        # (non-threading) thread.
+        lock = self.locktype()
+        def f():
+            lock.acquire()
+            lock.release()
+        n = len(threading.enumerate())
+        # We run many threads in the hope that existing threads ids won't
+        # be recycled.
+        self.Bunch(f, 15).wait_for_finished()
+        self.assertEqual(n, len(threading.enumerate()))
+
+    """
+    Tests for non-recursive, weak locks
+    (which can be acquired and released from different threads).
+    """
+
+    def test_reacquire(self):
+        # Lock needs to be released before re-acquiring.
+        lock = self.locktype()
+        phase = []
+        def f():
+            lock.acquire()
+            phase.append(None)
+            lock.acquire()
+            phase.append(None)
+        start_new_thread(f, ())
+        while len(phase) == 0:
+            _wait()
+        _wait()
+        self.assertEqual(len(phase), 1)
+        lock.release()
+        while len(phase) == 1:
+            _wait()
+        self.assertEqual(len(phase), 2)
+
+    def test_different_thread(self):
+        # Lock can be released from a different thread.
+        lock = self.locktype()
+        lock.acquire()
+        def f():
+            lock.release()
+        b = self.Bunch(f, 1)
+        b.wait_for_finished()
+        lock.acquire()
+        lock.release()
+
+    """
+    Tests for recursive locks.
+    """
+    def test_reacquire(self):
+        lock = self.locktype()
+        lock.acquire()
+        lock.acquire()
+        lock.release()
+        lock.acquire()
+        lock.release()
+        lock.release()
+
+    def test_release_unacquired(self):
+        # Cannot release an unacquired lock
+        lock = self.locktype()
+        self.assertRaises(RuntimeError, lock.release)
+        lock.acquire()
+        lock.acquire()
+        lock.release()
+        lock.acquire()
+        lock.release()
+        lock.release()
+        self.assertRaises(RuntimeError, lock.release)
+
+    def test_different_thread(self):
+        # Cannot release from a different thread
+        lock = self.locktype()
+        def f():
+            lock.acquire()
+        b = self.Bunch(f, 1, True)
+        try:
+            self.assertRaises(RuntimeError, lock.release)
+        finally:
+            b.do_finish()
+
+    def test__is_owned(self):
+        lock = self.locktype()
+        self.assertFalse(lock._is_owned())
+        lock.acquire()
+        self.assertTrue(lock._is_owned())
+        lock.acquire()
+        self.assertTrue(lock._is_owned())
+        result = []
+        def f():
+            result.append(lock._is_owned())
+        self.Bunch(f, 1).wait_for_finished()
+        self.assertFalse(result[0])
+        lock.release()
+        self.assertTrue(lock._is_owned())
+        lock.release()
+        self.assertFalse(lock._is_owned())
+
+
+
 if __name__ == '__main__':
     import os
     import unittest
